@@ -53,6 +53,8 @@
 #include "certtool-args.h"
 #include "certtool-common.h"
 
+#define MAX_HASH_SIZE 64
+
 static FILE *stdlog = NULL;
 
 static void print_crl_info(gnutls_x509_crl_t crl, FILE * out);
@@ -568,6 +570,18 @@ generate_certificate(gnutls_privkey_t * ret_key,
 			}
 		}
 
+		result = get_email_protection_status();
+		if (result) {
+			result =
+			    gnutls_x509_crt_set_key_purpose_oid
+			    (crt, GNUTLS_KP_EMAIL_PROTECTION, 0);
+			if (result < 0) {
+				fprintf(stderr, "key_kp: %s\n",
+					gnutls_strerror(result));
+				exit(1);
+			}
+		}
+
 		if (ca_status) {
 			result = get_cert_sign_status();
 			if (result)
@@ -1481,7 +1495,7 @@ void pgp_privkey_info(void)
 	gnutls_openpgp_privkey_t key;
 	unsigned char keyid[GNUTLS_OPENPGP_KEYID_SIZE];
 	size_t size;
-	int ret, i, subkeys, bits = 0;
+	int ret, i, subkeys;
 	gnutls_datum_t pem;
 	const char *cprint;
 
@@ -1555,8 +1569,6 @@ void pgp_privkey_info(void)
 				print_rsa_pkey(outfile, &m, &e, &d, &p, &q,
 					       &u, NULL, NULL,
 					       HAVE_OPT(CPRINT));
-
-			bits = m.size * 8;
 		} else if (ret == GNUTLS_PK_DSA) {
 			gnutls_datum_t p, q, g, y, x;
 
@@ -1575,8 +1587,6 @@ void pgp_privkey_info(void)
 			else
 				print_dsa_pkey(outfile, &x, &y, &p, &q, &g,
 					       HAVE_OPT(CPRINT));
-
-			bits = y.size * 8;
 		}
 
 		fprintf(outfile, "\n");
@@ -1615,21 +1625,8 @@ void pgp_privkey_info(void)
 				"Error in fingerprint calculation: %s\n",
 				gnutls_strerror(ret));
 		} else {
-			gnutls_datum_t art;
-
 			fprintf(outfile, "Fingerprint: %s\n",
 				raw_to_string(lbuffer, size));
-
-			ret =
-			    gnutls_random_art(GNUTLS_RANDOM_ART_OPENSSH,
-					      cprint, bits, lbuffer, size,
-					      &art);
-			if (ret >= 0) {
-				fprintf(outfile,
-					"Fingerprint's random art:\n%s\n\n",
-					art.data);
-				gnutls_free(art.data);
-			}
 		}
 	}
 
@@ -2079,6 +2076,18 @@ void generate_request(common_info_st * cinfo)
 			}
 		}
 
+		ret = get_email_protection_status();
+		if (ret) {
+			ret =
+			    gnutls_x509_crq_set_key_purpose_oid
+			    (crq, GNUTLS_KP_EMAIL_PROTECTION, 0);
+			if (ret < 0) {
+				fprintf(stderr, "key_kp: %s\n",
+					gnutls_strerror(ret));
+				exit(1);
+			}
+		}
+
 		ret = get_ipsec_ike_status();
 		if (ret) {
 			ret = gnutls_x509_crq_set_key_purpose_oid
@@ -2949,7 +2958,9 @@ void pkcs7_sign(common_info_st * cinfo, unsigned embed)
 	size_t size;
 	gnutls_datum_t data;
 	unsigned flags = 0;
-	gnutls_x509_crt_t signer;
+	gnutls_x509_crt_t *crts;
+	size_t crt_size;
+	size_t i;
 
 	if (ENABLED_OPT(P7_TIME))
 		flags |= GNUTLS_PKCS7_INCLUDE_TIME;
@@ -2971,18 +2982,27 @@ void pkcs7_sign(common_info_st * cinfo, unsigned embed)
 		exit(1);
 	}
 
-	signer = load_cert(1, cinfo);
+	crts = load_cert_list(1, &crt_size, cinfo);
 	key = load_private_key(1, cinfo);
 
 	if (embed)
 		flags |= GNUTLS_PKCS7_EMBED_DATA;
 
-	ret = gnutls_pkcs7_sign(pkcs7, signer, key, &data, NULL, NULL, get_dig(signer), flags);
+	ret = gnutls_pkcs7_sign(pkcs7, *crts, key, &data, NULL, NULL, get_dig(*crts), flags);
 	if (ret < 0) {
 		fprintf(stderr, "Error signing: %s\n", gnutls_strerror(ret));
 		exit(1);
 	}
 
+	for (i=1;i<crt_size;i++) {
+		ret = gnutls_pkcs7_set_crt(pkcs7, crts[i]);
+		if (ret < 0) {
+			fprintf(stderr, "Error adding cert: %s\n", gnutls_strerror(ret));
+			exit(1);
+		}
+	}
+
+
 	size = lbuffer_size;
 	ret =
 	    gnutls_pkcs7_export(pkcs7, outcert_format, lbuffer, &size);
@@ -2994,7 +3014,10 @@ void pkcs7_sign(common_info_st * cinfo, unsigned embed)
 	fwrite(lbuffer, 1, size, outfile);
 
 	gnutls_privkey_deinit(key);
-	gnutls_x509_crt_deinit(signer);
+	for (i=0;i<crt_size;i++) {
+		gnutls_x509_crt_deinit(crts[i]);
+	}
+	gnutls_free(crts);
 	gnutls_pkcs7_deinit(pkcs7);
 	exit(0);
 }
@@ -3761,15 +3784,17 @@ void pkcs7_info(common_info_st *cinfo, unsigned display_data)
 			exit(1);
 		}
 	} else {
-		ret = gnutls_pkcs7_print(pkcs7, GNUTLS_CRT_PRINT_FULL, &str);
-		if (ret < 0) {
-			fprintf(stderr, "printing error: %s\n",
-				gnutls_strerror(ret));
-			exit(1);
-		}
+		if (outcert_format == GNUTLS_X509_FMT_PEM) {
+			ret = gnutls_pkcs7_print(pkcs7, GNUTLS_CRT_PRINT_FULL, &str);
+			if (ret < 0) {
+				fprintf(stderr, "printing error: %s\n",
+					gnutls_strerror(ret));
+				exit(1);
+			}
 
-		fprintf(outfile, "%s", str.data);
-		gnutls_free(str.data);
+			fprintf(outfile, "%s", str.data);
+			gnutls_free(str.data);
+		}
 
 		size = lbuffer_size;
 		ret =
@@ -3993,8 +4018,8 @@ static
 void pubkey_keyid(common_info_st * cinfo)
 {
 	gnutls_pubkey_t pubkey;
-	uint8_t fpr[64];
-	char txt[256];
+	uint8_t fpr[MAX_HASH_SIZE];
+	char txt[MAX_HASH_SIZE*2+1];
 	int ret;
 	size_t size, fpr_size;
 	gnutls_datum_t tmp;
@@ -4011,7 +4036,7 @@ void pubkey_keyid(common_info_st * cinfo)
 	else if (default_dig == GNUTLS_DIG_SHA256)
 		flags = GNUTLS_KEYID_USE_SHA256;
 	else {
-		fprintf(stderr, "Cannot calculate key ID with the provided hash\n");
+		fprintf(stderr, "Cannot calculate key ID with the provided hash (use sha1, or sha256)\n");
 		exit(1);
 	}
 
@@ -4051,8 +4076,8 @@ void certificate_fpr(common_info_st * cinfo)
 	int ret = 0;
 	gnutls_datum_t pem, tmp;
 	unsigned int crt_num;
-	uint8_t fpr[32];
-	char txt[128];
+	uint8_t fpr[MAX_HASH_SIZE];
+	char txt[MAX_HASH_SIZE*2+1];
 	size_t fpr_size;
 
 	crt = load_cert(0, cinfo);
